#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <cblas.h>
#include <clapack.h>
#include "util.h"
#include "jacseq.h"

/*--- Calcul du residu :
  R = B - A * X
*/
int RESIDU(double *A, double *X, double *B, double *R, int Nblocs, int M){

  int blocdiag, bloc, shift, bloconline, blocsperline;
  
  /* Initialisation R <- B */
  cblas_dcopy(Nblocs * M, B, 1, R, 1);
  
  bloc = 0;
  /* Puis R <- R - A * X par bloc*/
  for (blocdiag = 0 ; blocdiag < Nblocs ; blocdiag++){
    if (Nblocs == 1){
      /* La matrice est en fait dense*/
      blocsperline = 1;
      shift=0;
    }
    else if (blocdiag == 0){
      /* Premier bloc de lignes */
      blocsperline = 2;
      shift=0;
    }
    else if (blocdiag == Nblocs - 1){
      /* Dernier bloc de lignes */
      blocsperline = 2;
      shift=-1;
    }
    else{
      /* Cas general */
      blocsperline = 3;
      shift=-1;
    }
    
    /*On parcourt les blocs et on met a jour R*/
    for (bloconline = 0; bloconline < blocsperline ; bloconline++){
      cblas_dgemv(CblasRowMajor, CblasNoTrans, M, M, -1.0, A+(bloc*M*M), M, 
		  X+((blocdiag + shift) * M), 1, 1.0, R+(blocdiag*M), 1);
      /*printf("bloc = %d ; shift = %d ; blocdiag = %d ; shift+blocdiag = %d; blocsperline = %d\n", bloc, shift, blocdiag, shift+blocdiag, blocsperline);*/
      bloc++;
      shift++;
    }
  }
  return 0;
}

/*-------------------------------------------------------------
C -Resolution du systeme A.X=B (de dimension NxN ) par 
C  la methode iterative de Jacobi par blocs. 
C -Soit A une matrice carree d'ordre N 
C      tridiagonale par blocs composee
C      de Nblocs blocs diagonaux de taille M (N = Nblocs*M)
C
C -EPS: definit la precision sur la solution recherchee.
C
C Le nombre d'iterations reellement effectuees Niter devra
C etre precise en parametre d'output.
C
C -Algorithme utilise: Version sequentielle  de Jacobi par bloc 
C -------------------------------------------------------------*/
/*----------
C Parametres: 
C ----------
C    Input  : A, B, N, M, Nblocs, Maxiter, EPS
C    Output : X, Niter
C
C    NB : ces parametres sont controles par l'utilisateur;
C    il est donc responsable de leur allocation et leur liberation.
C
C ------------------
C Variables locales
C ----------------------------------
C R(N): Vecteur residu a chaque iteration
C Y(M): Vecteur iterant supplementaire
C
C Pour la factorization LU des matrices Akk (Nblocs blocs diagonaux de A)
C    Akk = Pk*Lk*Uk , ou 
C      Pk est une matrice de permutation 
C      Lk et Uk sont respectivement des matrices triangulaires
C                  inferieures et superieures.
C   on introduit:
C     - D(M,M,Nblocs):  Matrice des blocs diagonaux factorises
C           D(1:M,1:M,k) (notee Dk) contiendra 
C           - dans un premier temps un copie de Akk, 
C           - puis Lk Uk associes a la factorisation de Akk
C     - IPIV (M, Nblocs) : 
C              IPIV(1..M, k) vecteur de permutations 
C              correspondant a une matrice de permutation Pk
C              et construit durant la factorization LU  de Akk
C Err : Norme max du residu
C Iter: Compteur d'iterations
C comm: Communicateur MPI (pour la version parallle)
C ----------------------------------*/
int jacseq (double *A, double *X, double *B, int N, int M, int Nblocs, double EPS, int Maxiter, int *Niter){

  double *R, *Y, *D;
  int *IPIV;
  double Err;
  int k,Iter;

  /*---------------------------------
    C On suppose que N = M * Nblocs
    C -------------------------------*/
  ASSERT(N = M * Nblocs);
  
  /*------------------------------------
    C Allocation des tableaux de travail
    C ---------------------------------*/
  R = APalloc(N * sizeof(double));
  ASSERT (R != NULL);
  Y = APalloc(M * sizeof(double));
  ASSERT (Y != NULL);
  D = APalloc(N * M * sizeof(double));
  ASSERT (D != NULL);
  IPIV = APalloc(N * sizeof(int));
  ASSERT (IPIV != NULL);
  
  /*-------------------------------------------
    C Initialisations de
    C     Iter = compteur d'iterations (K=0)
    C     X    = vecteur iterant (X(i)=0, i=1,N) 
    C     R    = vecteur residu  (R(i)=B(i), i=1,N)
    C     Err  = || R ||
    C -------------------------------------------*/
  Iter = 0;
  initarray(X, 0, N);
  /* R <- B */
  cblas_dcopy(N, B, 1, R, 1); /* NB : on aurait pu utiliser memcpy
				 ou derouler la boucle
				 for (i = 0; i < N; i++){R[i]=B[i];*/
  /*-- Calcul de la premiere norme de residu*/
  Err  = NORM(R,N);
  printf("Iteration %d : norme de B = %e\n",Iter ,Err);

  /*----------------------------------------------
    C Copie et 
    C Factorisation LU des Nblocs blocs diagonaux Aj
    C ----------------------------------------------*/
  for (k=0 ; k < Nblocs ; k++){
    /*-- copy de Akk dans Dk */
    cblas_dcopy(M * M, A+(3*k*M*M), 1, D+(k*M*M), 1);
    /* -- factorisation LU de Dk  (Dk = Pk*Lk*Uk) */
    clapack_dgetrf(CblasRowMajor, M, M, D+(k*M*M), M, IPIV+(k*M));
    /* -- Dk contient les matrices Lk et Uk. 
       -- IPIV (..:N,k) est le vecteur de permutation associe a Pk */
  }

  /*-----------------------
    C Boucle de convergence
    C -----------------------*/
  while ((Err > EPS) && (Iter < Maxiter)){ 
    /*---------------------------
      C Boucle sur les blocs de A
      C ------------------------*/
    for (k = 0 ; k < Nblocs ; k++){
      /* -- calcul de Y tel que Dk Y = Rk  
	 C ou Rk = R((k-1)*M+1, k*M))
	 C (en entree de DGETRS Y contient Rk)*/
      cblas_dcopy(M, R+(k*M), 1, Y, 1);
      clapack_dgetrs(CblasRowMajor, CblasNoTrans,
		     M, 1, D+(k*M*M), M,
		     IPIV+(k*M), Y, M);
      /* (en sortie de DGETRS Y contient la solution de 
	 C Dk Y = Rk )
	 C -- Calcul du nouvel iterant*/
      cblas_daxpy(M, 1.0, Y, 1,
		  X+(k*M), 1);
    }
    /*---------------------------------------
      C Calcul du nouveau residu et de sa norme
      C --------------------------------------*/
    RESIDU (A, X, B, R, Nblocs, M);
    Err = NORM(R,N);
    Iter=Iter+1;
    printf("Iteration %d : norme du residu = %e\n",Iter ,Err);
  }

  free(R);
  free(Y);
  free(D);    
  free(IPIV);
  *(Niter) = Iter;
  return 0;
}

